#include <iostream>
#include <fstream>
#include <utility> //for std::pair
#include <boost/graph/graphviz.hpp>

#include "pu.hh"
#include "vcg.hh"

template < class T >
class node_writer {

public:
  node_writer(T _t) : t(_t) {}
  void operator()(std::ostream & _out, vertex_id _v_id) const {
    lcc::Node u = get_lcc_node(_v_id, t);
    int time_stamp = get_time_stamp(_v_id, t);
    _out << "[label=\"" << lcc::opname(u->op) << "\\n"
	 << u->op << "\\n"
	 << _v_id << ":"
	 << time_stamp << "\"]";
  }
private:

  T t;	
};

template < class T >
class edge_writer {

public:
  edge_writer(T _t) : t(_t) {}
  void operator()(std::ostream & _out, edge_id _e_id) const {
    _out << "[label=\"" << edgetype(get_edge_type(_e_id, t)) << "\"]";
  }

private:
  T t;
};
	
PU::PU(lcc::Node p, Architecture *_architecture)
  : architecture(_architecture)
{
  lcc::Node q, q_start, q_end;
  static int time_stamp;
  std::map< int, lcc::Node > stamp_map;
  //  dependencies.clear();
	
  stamp_map.clear();
  current_dag = new DAG;
  dag_nodes_map[current_dag].clear();
  dags.push_back(current_dag);
	
  q_start = q_end = p;
  for(lcc::Node q = p; q; q = q->link) {
    stamp_map[time_stamp] = q;

    //if q alerady in, do not rebuild
    dag_nodes_map_t::iterator nodes_iter = dag_nodes_map.find(current_dag);
    assert(nodes_iter != dag_nodes_map.end());
    node_id_map_t::const_iterator v_iter = (nodes_iter->second).find(q);
    if(v_iter != (nodes_iter->second).end()) {
      time_stamp++;
      continue;
    }
    build_dag(q, time_stamp, *current_dag);

    //find INDIR_x->( )_y, y < x, then add edge TOP_y->INDIR_x
    //    assert(dependencies.empty());
    //    assert(deps_stack.empty());
    deps_t deps;
    deps.clear();
    find_deps(time_stamp, q, *current_dag, deps);

    //Add DDEP edges
    for(int i = 0; i < deps.size(); i++) {
      vertex_id top_y = get_vertex_id(stamp_map[ deps[i].first ],
				      *current_dag);
      //Add a new edge, if there is no already existing one
      bool edge_exists = false;
      out_edge_iter ei, ei_end;
      for(tie(ei, ei_end) = out_edges(top_y, *current_dag); 
	  ei != ei_end; ++ei) {
	if(target(*ei, *current_dag) == deps[i].second) {
	  edge_exists = true;
	  break;
	}
      }
      if((!edge_exists) && (top_y != deps[i].second)) {
	add_edge(top_y, deps[i].second, DDEP, *current_dag);
      }
    }

//     assert(deps_stack.empty());
//     for(int i = 0; i < dependencies.size(); i++) {
//       deps * d = dependencies[i];
//       for(int j = 0; j < d->time_stamps.size(); j++) {
// 	vertex_id top_y = 
// 	  get_vertex_id(stamp_map[d->time_stamps[j]],
// 			*current_dag);
// 	//Add a new edge, if there is no already existing one
// 	out_edge_iter ei, ei_end;
// 	bool b_exist = false;
// 	for(tie(ei, ei_end) = out_edges(top_y, *current_dag); 
// 	    ei != ei_end; ++ei) {
// 	  if(target(*ei, *current_dag) == d->indir) {
// 	    b_exist = true;
// 	    break;
// 	  }
// 	}
// 	if(!b_exist) {
// 	  add_edge(top_y, d->indir, DDEP, *current_dag);
// 	}
//       }
//       delete d;
//     }
//     dependencies.clear();
    

    time_stamp++;

    switch(generic(q->op)) {
      //Nodes that ends BB
    case lcc::EQ: case lcc::GE: case lcc::GT: 
    case lcc::LT: case lcc::NE:	case lcc::CALL: 
    case lcc::RET: case lcc::JUMP: case lcc::LABEL:
      {
	lcc::Node r;
	//Add link edges between q_start and q_end
	q_end = q;
	for(r = q_start; r != q_end; r = r->link) {
										
	  vertex_id p_id = get_vertex_id(r, *current_dag),
	    q_id = get_vertex_id(r->link, *current_dag);
	  add_edge(p_id, q_id, LINK, *current_dag);
	}
	if(q->link) {
	  current_dag = new DAG;
	  dag_nodes_map[current_dag].clear();
	  dags.push_back(current_dag);
	  q_start = q_end = q->link;
	}
	q_start = q_end;
      }
    }
  }
}

PU::~PU()
{
}

void PU::find_deps(int _time_stamp, lcc::Node _p, DAG & _dag, deps_t & _deps,
		   vertex_id _predecessor)
{
  if(_p) {
    vertex_id p_id = get_vertex_id(_p, _dag);
    int p_ts = get_time_stamp(p_id, _dag);
    if((p_ts < _time_stamp) && (_predecessor != (vertex_id)-1)) {
      _deps.push_back(std::make_pair(p_ts, _predecessor));
    } else {
      for(int i = 0; i < NELEMS(_p->kids); i++) {
	find_deps(_time_stamp, _p->kids[i], _dag, _deps, p_id);
      }
    }
  }
}

// void PU::find_deps(int _time_stamp, lcc::Node _p, DAG & _dag)
// {
//   if(_p) {
//     vertex_id p_id = get_vertex_id(_p, _dag);
//     int p_ts = get_time_stamp(p_id, _dag);
//     if(p_ts != _time_stamp) {
//       assert(p_ts < _time_stamp);
//       if(!deps_stack.empty()) {
// 	//There are dependecies
// 	deps *tmp_deps = deps_stack.top();
// 	tmp_deps->time_stamps.push_back(p_ts);
//       } //otherwise the dependecies are satified by existing edges
//     } else {
//       if(generic(_p->op) == lcc::INDIR) {
// 	deps *new_deps = new deps;
// 	new_deps->indir = p_id;
// 	deps_stack.push(new_deps);
// 	assert(_p->kids[0]); //INDIR has only on child
// 	find_deps(_time_stamp, _p->kids[0], _dag);
// 	deps *top_deps = deps_stack.top();
// 	deps_stack.pop();
// 	if(top_deps->time_stamps.empty()) {
// 	  delete top_deps;
// 	} else {
// 	  dependencies.push_back(top_deps);
// 	}
//       } else {
// 	for(int i = 0; i < NELEMS(_p->kids); i++) {
// 	  find_deps(_time_stamp, _p->kids[i], _dag);
// 	}
//       }
//     }
//   }
// }

void PU::print(bool _vcg, bool _dot)
{
  char file_name[12];
  static int dag_number;
  DAGS::iterator d_iter = dags.begin();
  for(; d_iter != dags.end(); ++d_iter)
    {
      DAG *dag = *d_iter;
      node_writer< DAG & > nw(*dag);
      edge_writer< DAG & > ew(*dag);

      //DOT
      if(_dot) {
	sprintf(file_name, "%s%.2d%s", "dag", dag_number, ".dot");
	std::ofstream dot_file(file_name);
	boost::write_graphviz(dot_file, *dag, nw, ew);
	dot_file.close();
      }

      //VCG
      if(_vcg) {
	sprintf(file_name, "%s%.2d%s", "dag", dag_number, ".vcg");
	std::ofstream vcg_file(file_name);
	VCG(*dag, vcg_file);
	vcg_file.close();
      }
      dag_number++;
    }
}

void PU::build_dag(lcc::Node _p, int _time_stamp, DAG & _dag)
{
  if(_p) {
    dag_nodes_map_t::iterator nodes_iter = dag_nodes_map.find(& _dag);
    assert(nodes_iter != dag_nodes_map.end());
    node_id_map_t::const_iterator v_iter = (nodes_iter->second).find(_p);
    if(v_iter == (nodes_iter->second).end()) {
      //Node not yet in the current DAG

      if(architecture->remove_op(_p->op)) { //Node to be removed
	assert(_p->kids[0] && !_p->kids[1]);
	build_dag(_p->kids[0], _time_stamp, _dag);
	node_id_map_t::const_iterator v_iter_prime = (nodes_iter->second).find(_p->kids[0]);
	assert(v_iter_prime != (nodes_iter->second).end());
	(nodes_iter->second)[_p] = v_iter_prime->second;
      } else {
	vertex_id u = add_vertex(_p, _time_stamp, _dag);
	(nodes_iter->second)[_p] = u;
	for (int i = 0; i < NELEMS(_p->kids); i++) {
	  build_dag(_p->kids[i], _time_stamp, _dag);
	  if(_p->kids[i]) {
	    vertex_id p_id = get_vertex_id(_p, _dag),
	      q_id = get_vertex_id(_p->kids[i], _dag);
	    add_edge(q_id, p_id, (i == 0 ? KID0 : KID1), _dag);
	  }
	}
      }
    }
  }
}

vertex_id PU::get_vertex_id(lcc::Node _p, DAG & _dag)
{
  dag_nodes_map_t::iterator nodes_iter = dag_nodes_map.find(& _dag);
  assert(nodes_iter != dag_nodes_map.end());
  node_id_map_t::const_iterator v_iter = (nodes_iter->second).find(_p);
  assert(v_iter != (nodes_iter->second).end());
  return (nodes_iter->second)[_p];
}
